dojo.provide('com.ibm.relm.ple.web.ui.TreeSelection'); //$NON-NLS-1$

(function(){

var selectedSites_ = [];

function itemPathToIdPath(itemPath, store) {
	return dojo.map(itemPath, function(item) {
		return store.getValue(item, 'pathId'); //$NON-NLS-1$
	});
}

/**
 * A utility method to help update the current ID path (used in the address bar)
 * after a check out or other operation that may modify an item version.
 * This to be used as a method on a `site` object (it is added in .newSitesFromItemPaths).
 *
 * When called on a site, it will create a path to an item version
 * at the same tree depth. That is, it will take the parent ID path
 * of the site, and add the pathId of a new item version. This is
 * used when refreshing after some modify operation. In many cases the
 * resulting path will be the same since "part equivalence IDs"
 * should not change during a check out. This is used for
 * historical reasons and since it's not clear if there are still
 * cases where it is truly needed.
 * 	
 * @param  {product} otherProduct the other product to get the pathId from to construct
 *                                	the full new path.
 * @return {array}              an ID path similar to the one on the starting site, but
 *                                 with a potentially different leaf if otherProduct
 *                                 had a different pathId.
 */
function idPathForOtherProduct(otherProduct) {
	var otherIdPath = this.idPath.slice(0, -1);
	otherIdPath.push(otherProduct.pathId);
	return otherProduct;
}

/**
 * This is a global location to track the currently selected item or items
 * in the product tree. This was added to help unify the previous myriad ways
 * the selection was being tracked or looked up.
 * The selection is tracked here for several reasons:
 * 
 * 	-	Decouples code from needing to reference the actual tree widget, which
 * 		is actually two trees, and may change t a grid at some point.
 * 	-	Simpler handling of the case where the tree hasn't loaded yet, but a widget
 * 		needs to know if there is a selection (there isn't) or not.
 * 	-	Normalizes the paths into just the parts most code is interested in (the selected product and parent product and maybe link)
 * 		as opposed to using the paths directly.
 * 	- 	Simplifies the selection change event since it's easy for other widgets to reach into the
 *  	selection after the event instead of counting on many parameters being passed to the callback.
 *
 * This class uses the term "site" to refer to the objects used to store
 * selections. A site is just an object with three proerties: "parentProduct",
 * "product", and "link". Only "product" is guaranteed to be non-null. "link"
 * is only set if the selected item is an artifact (link). "parentProduct" is
 * the parent of child products, but is null for top-level products. Note that
 * a "link" is always under a "product", but the product may or may not have a
 * "parentProduct" if it's top-level or not. Don't be confused, the parent of
 * the link is "product" since links are just URL properties. These properties
 * are all item objects with the properties of their resources.
 *
 * Sites also have a "idPath" property to identify their full path in the tree
 * with string IDs in an array.
 * 
 * A "site" object is convenient since it describes the entire context of a
 * selection that may be needed for checkout and modification.
 */
dojo.mixin(com.ibm.relm.ple.web.ui.TreeSelection, {

	/**
	 * The data store that that the tree (or future widget) is using
	 * to access items in it. Any called that reads the selected items
	 * from this object must access their properties via the store methods.
	 * 
	 * This feels pretty fragile, since the tree must set this once it's loaded.
	 * Once set, this never changes.
	 * 
	 * @type {data.store}
	 */
	store: null,

	setSelectedSitesFromItemPaths: function(itemPaths) {
		var newSelection = this.newSitesFromItemPaths(itemPaths);
		this.setSelectedSites(newSelection);
	},

	/**
	 * Creates an array of "site" object structures. This doesn't
	 * actually set them as the global selection, it just creates the objects.
	 * This is used internally when setting them, but can also be called
	 * for creating artificial "selections" for drag and drop.
	 * @param  {[type]} itemPaths [description]
	 * @return {[type]}           [description]
	 */
	newSitesFromItemPaths: function(itemPaths) {
		var selectedSites = [];

		dojo.forEach(itemPaths, function(path) {
			
			if (path.length === 0) return;

			var product = path[path.length - 1];
			var type = this.store.getValue(product, 'type'); //$NON-NLS-1$
			var link = null;

			if (type === 'product') { //$NON-NLS-1$
				var parentProduct = path[path.length - 2];
			}
			else {
				// Not a product, it's an artifact link or hyperlink.
				parentProduct = path[path.length - 3];
				product = path[path.length - 2];
				link = path[path.length - 1];
			}

			var site = {
				parentProduct: parentProduct,
				product: product,
				link: link,
				idPath: itemPathToIdPath(path, this.store),
				itemPath: path,
				idPathForOtherProduct: idPathForOtherProduct
			};

			selectedSites.push(site);
		}, this);

		return selectedSites;
	},

	/**
	 * Takes a new array of selected objects and sets them as the current
	 * selection. Notifies any listeners.
	 * @param {array} newSelection array of objects with .parentProduct and .product fields on them.
	 * Array of 1 if there is only one selected item, array of more if multi-select was used.
	 * Can be emtpy array for no selection.
	 */
	setSelectedSites: function(newSelection) {
		selectedSites_ = newSelection.slice(0);
		this.onSelectionChange();
	},

	/**
	 * Clears all the selected paths. The selection will be an empty array.
	 */
	clearSelectedSites: function() {
		this.setSelectedSites([]);
	},

	/**
	 * Gets all the selected items, assuming multi-select was used
	 * and there is more than one. See .getSelected for details.
	 * @return {array}          array of 0, 1, or more selected objects.
	 */
	getAllSelectedSites: function() {
		return selectedSites_.slice(0);
	},

	/**
	 * Gets the main selected object. If multi select was used, this is the last selected object,
	 * and the current focused object.
	 * 
	 * @return {object}          an object with a .product property for lowest product item item.
	 *                           This can be the selected product if a product is selected, or
	 *                           the product with a link on it, if a link is selected.
	 *                           If the .product was not a top-level item, the result will have a .parentProduct
	 *                           field with the parentProduct item as well.
	 *                           If the selected item is a link, a .link property holds the link item.
	 *                           Will have a .idPath property, which is an array of tree IDs to the selected
	 *                           item, excluding the invisible root node.
	 */
	getSelectedSite: function() {
		if (selectedSites_.length > 0) {
			return selectedSites_[selectedSites_.length - 1];
		}
		else {
			return null;
		}
	},

	/**
	 * Convenience function to return the 'type' of the currently selected item,
	 * that is, the type of the .product or .link (if a link is selected) that would be returned by .getSelected.
	 * @return {string} null if no selection, otherwise the type of the selected leaf.
	 */
	getSelectedType: function() {
		if (selectedSites_.length === 0 || !this.store) {
			return null;
		}

		var leaf = selectedSites_[selectedSites_.length - 1];
		return this.store.getValue(leaf.link || leaf.product, 'type'); //$NON-NLS-1$
	},

	/**
	 * For all selected nodes, check that they are all product types (no links.)
	 * @return {boolean} true if all selected nodes are products.
	 */
	areAllSelectedProducts: function() {
		return dojo.every(selectedSites_, function(selected) {
			return !selected.link;
		}, this);
	},

	/**
	 * Callback functiom. Use dojo.connect to listen to changes to the
	 * selection. No parameters are passed to the callback, so interested parties
	 * must use other TreeSelection methods to inpsect the selection.
	 */
	onSelectionChange: function() {
		// no-op callback
	}

});	

})();