/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2010. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
dojo.provide('com.ibm.relm.ple.web.ui.tree.ProductStore'); //$NON-NLS-1$

dojo.require('dojo.data.api.Read'); //$NON-NLS-1$
dojo.require('dojo.data.api.Write'); //$NON-NLS-1$
dojo.require('dojo.data.api.Notification'); //$NON-NLS-1$
dojo.require('dojo.data.api.Identity'); //$NON-NLS-1$

(function(){

	var DO_NOT_SET = {
		'treeId': true, //$NON-NLS-1$
		'$ref': true //$NON-NLS-1$
	};

	/**
	 * When caching items, we to send notifications about any changes. But
	 * sending them in parallel causes callbacks while the cache is still being updated.
	 * This class defers all those until later. They are just added to a queue, then all
	 * dispatched in a setTimeout later on.
	 * 
	 * @param {Store} forStore the store to send the notifications from. Normally ProductStore.
	 */
	function PendingNotifications(forStore) {
		this.store = forStore;
		this.queuedNotifications = [];
	}

	PendingNotifications.prototype.queueOnSet = function(item, attribute, oldValue, newValue) {
		if (DO_NOT_SET[attribute]) return;

		if (oldValue != newValue)	{
			// console.log('Will update ' + item.name, attribute, oldValue, newValue);
			this.queuedNotifications.push({
				method: 'onSet', //$NON-NLS-1$
				args: [item, attribute, oldValue, newValue]
			});
		}
	};

	/**
	 * Compares two items and queues an onSet event for each changed or deleted attribute.
	 * @param  {Item} oldItem the old item from the store.
	 * @param  {Item} newItem the new item being added to the store with updated attributes (maybe).
	 */
	PendingNotifications.prototype.queueOnSetForItems = function(oldItem, newItem) {
		// TODO: check etag for differences to skip this

		for (var attribute in newItem) {
			var oldValue = oldItem[attribute];
			var newValue = newItem[attribute];
			if (oldValue != newValue) {
				this.queueOnSet(newItem, attribute, oldValue, newValue);
			}
		}

		for (var oldAttribute in oldItem) {
			if (newItem[oldAttribute] === undefined) {
				this.queueOnSet(newItem, oldAttribute, oldItem[oldAttribute], null);
			}
		}
	};

	PendingNotifications.prototype.queueOnNew = function(item) {
		this.queuedNotifications.push({
			method: 'onNew', //$NON-NLS-1$
			args: [item]
		});
	};

	PendingNotifications.prototype.send = function() {

		if (this.queuedNotifications.length === 0) return;
		// console.log('SEND', this.queuedNotifications);

		var queuedNotificationsCopy = [].concat(this.queuedNotifications);
		this.queuedNotifications = [];
		
		setTimeout(dojo.hitch(this, function() {
			// Send this on the next loop so that we're doing updating the cache before things reach back
			// in after being notified. Not sure if this is needed or not.
			for (var i = 0; i < queuedNotificationsCopy.length; i++) {
				var pending = queuedNotificationsCopy[i];
				var method = this.store[pending.method];
				method.apply(this.store, pending.args);
			}

		}), 1);

	};

	/**
	 * A custom store used to hold products for the product tree. Supports lazy loading
	 * in a manner similar to JsonRestStore (uses $ref to designate partially loadeded items
	 * and uses children: true before children are loaded.)
	 * This custom store is used to tightly control the unque refresh behaviour needed by
	 * the product tree.
	 */
	dojo.declare(
		'com.ibm.relm.ple.web.ui.tree.ProductStore', //$NON-NLS-1$
		[ dojo.data.api.Read, dojo.data.api.Write, dojo.data.api.Identity, dojo.data.api.Notification ],
		{
			constructor: function(args)
			{
				this._url = args.target;

				this._treeIdsToItems = {};
			},

			// Callbacks from Notification API
			onSet: function() {},
			onNew: function() {},
			onDelete: 	function() {},

			getIdentity: function(item) {
				if (!item) console.trace('Item is null'); //$NON-NLS-1$
				return item.treeId || item.$ref;
			},

			getAllItems: function() {
				var all = [];

				for (var treeId in this._treeIdsToItems) {
					all.push(this._treeIdsToItems[treeId]);
				}

				return all;
			},

			hasAttribute: function(item, attribute) {
				item = this._getCachedItem(item);
				// Sometimes this is `undefined`. Not sure why, but it always means false.
				if (!attribute) return false;

				var value = item[attribute];

				if (!value) {
					// No value (or `false`) means NO
					return false;
				}
				else {
					// Empty array is essentially no. This works
					// around a weird timing error after refresh, where artifacts look
					// like they have a `children` value of []. That's wrong, but needs to return NO.
					// Feels like a hack... may need to debug this later.
					if (value.length === 0) return false;
				}

				return true;
			},

			setValue: function(item, attribute, newValue) {
				if (DO_NOT_SET[attribute]) {
					throw new Error('setting illegal attribute: ' + attribute); //$NON-NLS-1$
				}

				var treeId = this.getIdentity(item);
				var otherItem = this._treeIdsToItems[treeId];

				if (!otherItem) {
					console.trace();
					return;
				}

				var oldValue = otherItem[attribute];
				if (oldValue != newValue) {
					otherItem[attribute] = newValue;
					this.onSet(otherItem, attribute, oldValue, newValue);
				}
			},

			setValueForAllWithSameUri: function(item, attribute, newValue) {
				if (DO_NOT_SET[attribute]) {
					throw new Error('setting illegal attribute: ' + attribute); //$NON-NLS-1$
				}

				// Items use a unique identity (treeId) for every location
				// they are in the product tree. But, we want to actually 
				// update all items with a matching (itemVersion)uri,
				// since we know they are the same item versions, and actually look
				// the same.
				var itemVersionUri = item.uri;

				for (var treeId in this._treeIdsToItems) {
					var otherItem = this._treeIdsToItems[treeId];

					if (!otherItem) {
						console.trace();
						return;
					}

					if (otherItem.uri === itemVersionUri) {
						var oldValue = otherItem[attribute];
						if (oldValue != newValue) {
							otherItem[attribute] = newValue;
							this.onSet(otherItem, attribute, oldValue, newValue);
						}
					}
				}
			},

			newItem: function(item, parentInfo) {
				var identity = this.getIdentity(item);
				if (!identity) {
					console.trace(item, parentInfo);
					throw new Error('No ID for new item'); //$NON-NLS-1$
				}
				if (this._treeIdsToItems[identity]) {
					throw new Error('Item already exists in store: ' + identity); //$NON-NLS-1$
				}

				this._treeIdsToItems[identity] = item;


				// Do NOT pass the parentInfo here. See http://dojotoolkit.org/reference-guide/1.6/dojo/data/api/Notification.html which says
				// "Note that if this function is invoked with a value for
				// parentInfo, then onSet is not invoked stating the attribute
				// of the parent item was modified. This is to avoid getting
				// two notification events occurring when a new item with a
				// parent is created."
				// Also, there seems to be a bug in TreeStoreModel that
				// incorrectly looks for parentInfo.item (not .parent) which
				// causes problems... not passing it works fine.
				this.onNew(item);

				if (parentInfo && parentInfo.parent) {
					var children = this.getValues(parentInfo.parent, parentInfo.attribute);
					children = children.concat([item]);
					this.setValues(parentInfo.parent, parentInfo.attribute, children);
				}
			},

			setValues: function(item, attribute, newValue) {
				if (!dojo.isArray(newValue)) {
					newValue = [newValue];
				}
				this.setValue(item, attribute, newValue);
			},

			getValue: function(item, attribute) {
				item =  this._getCachedItem(item);
				return item[attribute];
			},

			_loadItemByUriAndCache: function(args) {
				jazz.client.xhrGet({
					handleAs: 'json', //$NON-NLS-1$
					url: this._url,
					content: {
						itemVersionUri: args.itemVersionUri,
						viewMemberUri: args.viewMemberUri
					},

					load: dojo.hitch(this, function (newItem) {
						// The server code 
						var cachedItem = this._cacheItems([newItem])[0];

						if (args.onItem) {
							dojo.hitch(args.scope, args.onItem)(cachedItem);
						}
					}),

					error: function(error) {
						console.error(error);
						args.onError && dojo.hitch(args.scope, args.onError)(error);
					}
				});
			},

			loadItem: function(args) {
				var existingItem = args.item;

				if (this.isItemLoaded(existingItem)) {
					if (args.onItem) {
						dojo.hitch(args.scope, args.onItem)(existingItem);
					}
					return;
				}

				this._loadItemByUriAndCache({
					itemVersionUri: existingItem.uri,
					viewMemberUri: existingItem['pd:part'], //$NON-NLS-1$
					onItem: function(newItem) {
						if (newItem.$ref) {
							throw new Error("Loaded item is still a $reference."); //$NON-NLS-1$
						}

						if (args.onItem) {
							dojo.hitch(args.scope, args.onItem)(newItem);
						}
					},
					onError: args.onError
				});
			},

			_getCachedItem: function(item) {
				var existingItem = this._treeIdsToItems[this.getIdentity(item)];
				
				if (!existingItem) {
					// Root item is fake, and not in cache. Use directly
					if (item.root) return item;

					console.error('Getting non-cached item', item, this._treeIdsToItems); //$NON-NLS-1$
					// console.trace();
					return item;
				}

				return existingItem;
			},

			getValues: function(item, attribute) {
				item =  this._getCachedItem(item);
				var value = item[attribute];
				if (!value) return [];

				if (dojo.isArray(value)) {
					// This is likely only the 'children' attribute.
					return value;
				}
				else {
					// TODO: should try to load the value 'sync' style here,
					// but normally we shouldn't get here since a properly
					// behaving tree model will always check that an item is
					// loaded before using it.

					if (attribute == 'children' && value === true) { //$NON-NLS-1$
						console.warn('Looking up real children of a partially loaded item ($ref item). Will return empty children.'); //$NON-NLS-1$
						return [];
					}
					else {
						return [value];
					}
				}
			},

			getLabel: function(item) {
				return this._getCachedItem(item).name;
			},

			isItemLoaded: function(item) {
				item =  this._getCachedItem(item);
				return !item.$ref && item.children !== true;
			},

			isItem: function(item) {
				var id = this.getIdentity(item);
				if (!id) return false;

				return !!this._treeIdsToItems[id];
			},

			/**
			 * Takes children (normally $refs) from a JSON call, and swaps out the instances
			 * of the child JSON objects with the equivalent ones in the cache.
			 * The result should be that there is only ever one instance of each
			 * item with the same treeId and they are all in the cache.
			 * 
			 * @param  {array of object} children the child to replace. This array will be modified in place.
			 */
			_swapChildRefsToCachedItems: function(children) {
				for (var i = 0; i < children.length; i++) {
						var child = children[i];
						var treeId = this.getIdentity(child);

						if (treeId) {
							var cachedChild = this._treeIdsToItems[treeId];

							if (cachedChild) {
								// console.debug('Swapping a child: ', child,  cachedChild);
								// There's already a cached item for this, so use that instead.
								children.splice(i, 1, cachedChild);
							}
						}
					}
			},


			/**
			 * Adds items to the local cache so they can be reused later.
			 * Will save  if they are there. Will replace  with real items
			 * once available. When loading a new item, it is assumed the children
			 * should be cached as well. If the children are already in the cache, then
			 * the parent's children reference is updated to point to the cached items.
			 * 
			 * @param  {array} items               the new items to cache.
			 * @param  {object} alreadySeenTreeIds  hash of already seen IDs to prevent infinite loops. Can start a null.
			 * @param  {PendingNotifications} 	queuedNotifications the current pending notifications. Starts as null, but a
			 *instance is passed down recursively.
			 */
			_cacheItems: function(items, alreadySeenTreeIds, queuedNotifications) {
				// Track all the items being cached this time to prevent
				// following circular dependencies between parents and
				// children
				alreadySeenTreeIds = alreadySeenTreeIds || {};

				var shouldSendPendingNotification = false;
				
				if (!queuedNotifications) {
					// None were sent in, so this is the top level call (not down in recursion).
					// When it finishes, send the notifications.
					shouldSendPendingNotification = true;
					queuedNotifications = new PendingNotifications(this);
				}

				var cachedItems = [];

				dojo.forEach(items, function(item) {
					var treeId = this.getIdentity(item);

					if (!treeId) {
						console.error('Missing identity: ', item); //$NON-NLS-1$
						throw new Error('missing identity: ' + item.name); //$NON-NLS-1$
					}

					if (alreadySeenTreeIds[treeId]) {
						console.debug('Skipping seen treeId:', treeId); //$NON-NLS-1$
						return;
					}
					alreadySeenTreeIds[treeId] = true;

					var existingItem = this._treeIdsToItems[treeId];

					if (!existingItem) {
						this._treeIdsToItems[treeId] = item;

						// TODO: do we ever need this during a cache???
						// I don't think so since this is only important if we were adding a new item that's has
						// a parent. But parent's are updated on their own already since they changed, so children
						// don't need to notify parents.
						// queuedNotifications.queueOnNew(item);
					}
					else
					{
						// console.debug('Existing', existingItem);

						// Don't try to recache the same object. Note that this is an idenity test and probably doesn't
						// work. Should look at etags.
						if (existingItem == item) {
							return;
						}

						if (existingItem.$ref || !item.$ref) {
							// If old item is a ref or the new one isn't, then the new one is better in every way.
							// We can just replace as is
							// console.log('Updating with new item', item);
							this._treeIdsToItems[treeId] = item;
						}
						else {
							// A new reference. It's children info is less complete, so don't lose what's already cached.
							// Make this reference look like a loaded item with the currently cached children.
							delete item.$ref;
							item.treeId = existingItem.treeId;

							if (!item.children || item.children.length === 0) {
								item.children = [];
							}
							else {
								item.children = existingItem.children;
							}
						}

						// Send notifications after we're done caching everything.
						queuedNotifications.queueOnSetForItems(existingItem, item);
					}

					cachedItems.push(item);

					if (item.children && dojo.isArray(item.children) && item.children.length > 0) {
						this._swapChildRefsToCachedItems(item.children);
						// If there are any new children that aren't in the cache yet, add them npow
						this._cacheItems(item.children, alreadySeenTreeIds, queuedNotifications);
					}
				}, this);
				
				if (shouldSendPendingNotification) queuedNotifications.send();
				return cachedItems;
			},

			/**
			 * Occasionally we check with the tree's to see what items they actually have loaded.
			 * There may be things in the cache that aren't in the tree anymore, so we can remove them
			 * from the cache so it doesn't grow forever.
			 * 
			 * @param  {set} inUseItems set of tree IDs that are still in use and should not be purged.
			 */
			purgeUnusedItems: function(inUseItems) {
				for (var treeId in this._treeIdsToItems) {
					if (!inUseItems[treeId]) {
						console.debug('Purging old item', this._treeIdsToItems[treeId]); //$NON-NLS-1$
						delete this._treeIdsToItems[treeId];
					}
				}
			},

			fetchItemByIdentity: function(args) {
				var treeId = args.identity;

				if (this._treeIdsToItems[treeId]) {
					if (args.onItem) dojo.hitch(args.scope, args.onItem)(this._treeIdsToItems[treeId]);
					return;
				}
				else {
					// tree ID will be 'itemVersionUri viewMemberUri' for children (separated with a space),
					//  and just 'itemVersionUri' for top level items. Either way, splitting
					//  on a space should yield an array or 1 or 2 items with item [0] always
					//  being the itemVersionUri
					
					// FIXME: this assumes identity is always itemVersionUri, but could be viewMemberUri
					args.itemVersionUri = args.identity;
					this._loadItemByUriAndCache(args);
				}
			},

			getAllItemsByUri: function() {
				var all = {};

				for (var treeId in this._treeIdsToItems) {
					var item = this._treeIdsToItems[treeId];
					all[item.uri] = item;
				}

				return all;
			},

			fetch: function(args) {
				var content = {};

				if (args.query) {
					content = args.query;
				}

				jazz.client.xhrGet({
					handleAs: 'json', //$NON-NLS-1$
					url: this._url,
					content: content,

					load: dojo.hitch(this, function (newItems) {
						var cachedItems = this._cacheItems(newItems);

						args.onBegin && args.onBegin(cachedItems.length, args);

						if (args.onItem) {
							dojo.forEach(cachedItems, function(item) {
								args.onItem(item);
							});
						}

						args.onComplete && args.onComplete(cachedItems, args);
					}),

					error: function(error) {
						console.error(error);
						args.onError && dojo.hitch(args.scope, args.onError)(error);
					}
				});
			},

			getFeatures: function() {
				return {
					'dojo.data.api.Read': true, //$NON-NLS-1$
					'dojo.data.api.Identity': true, //$NON-NLS-1$
					'dojo.data.api.Notification': true //$NON-NLS-1$
				};
			}
		}
	);

})();