dojo.provide('com.ibm.relm.ple.web.ui.tree.TreeRefresh'); //$NON-NLS-1$

dojo.require('com.ibm.relm.ple.web.ui.Goto'); //$NON-NLS-1$
dojo.require("net.jazz.ajax.ui.PlatformUI"); //$NON-NLS-1$
dojo.require('com.ibm.relm.common.web.StatusMessage'); //$NON-NLS-1$

dojo.require('dojo.i18n'); //$NON-NLS-1$
dojo.requireLocalization('com.ibm.relm.ple.web.ui.tree', 'TreeRefreshMessages'); //$NON-NLS-1$ //$NON-NLS-2$

/**
 * Simplifies refreshing the tree and selecting the proper node afterwards.
 * Exposes static functions that decouple calling code from the exact implementation of the
 * product tree (which has changed several times) and provides and simple interface
 * for refreshing.
 */
(function(){

var Goto = com.ibm.relm.ple.web.ui.Goto;
var StatusMessage = com.ibm.relm.common.web.StatusMessage;

/**
 * Returns the parent path up to the last path component.
 * @param  {array} path the path
 * @return {array}      parent path.
 */
function parentPath(path) {
	return path.slice(0, path.length - 1);
}

/**
 * The single known product tree. This gets reset by the sidebar after a full refresh,
 * but is generally always set.
 */
var tree_ = null;

var messages_ = dojo.i18n.getLocalization('com.ibm.relm.ple.web.ui.tree', 'TreeRefreshMessages'); //$NON-NLS-1$ //$NON-NLS-2$

dojo.mixin(com.ibm.relm.ple.web.ui.tree.TreeRefresh, {

	setTree: function(newTree) {
		tree_ = newTree;
	},

	/**
	 * Gets the last set tree, which should be the visible product tree
	 * except during a hard refresh when it's not yet set. 
	 * This feels like a bit of a hack exposing this here and was only
	 * added as a convenience for drag and drop.
	 * @return {AltProductTree} the current product tree or null if not
	 * set yet.
	 */
	getTree: function() {
		return tree_;
	},

	/**
	 * Refreshes the product tree. All visible nodes will be updated with the latest changes.
	 * Collapsed nodes may be uncached and will need to lazy load again. See `args` for 
	 * operating on the refreshed tree.
	 * 	
	 * @param  {object} args args.selectPath to open the tree and select a product
	 * after the refresh. A path is a forward-slash separated string of `pathId`s starting
	 * from a root product.
	 * 		args.selectPathPath is like selectPath, but the parent (one from the leaf of the path) will be selected.
	 * 		This is used when adding links, which we don't want to select--we select the parent product. This is
	 * 		mutually exclusive with selectPath.
	 * 		args.flashPath to visually flash the selected node provided in .selectPath, for example,
	 * 		after a product create. If args.selectParentPath is used, then the leaf node of the path (usually a link)
	 * 		will stull flash, even though the parent is what is selected.
	 * 		args.openPath to open the selected node (normally only its parents are opened). Mutually exclusive with flashPath.
	 * @return {dojo.Deferred}      callbacks will fire after the tree is refreshed. Callers
	 * usually should use the know when to hide progress indicators or close a dialog.
	 */
	refresh: function(args) {
		dojo.publish("ple.product_tree_hide_message"); //$NON-NLS-1$
		
		args = args || {};
		
		if (!tree_) {
			console.warn('Product tree doesn\'t exist to refresh'); //$NON-NLS-1$
			return false;
		}

		var refreshingMessage = StatusMessage.show(messages_.refreshingProducts);

		var deferredRefresh = tree_.refresh();

		if (args.selectPath || args.selectParentPath) {

			var pathToSelect = args.selectParentPath ? parentPath(args.selectParentPath) : args.selectPath;

			deferredRefresh.then(function() {
				Goto.gotoProductTreePath(pathToSelect);

				if (args.openPath) {
					var pathToOpen = args.selectParentPath || args.selectPath;
					tree_.getNodeAtIdPath(pathToOpen).open();
				}
				else if (args.flashPath) {
					var flashArgs = {
						// Match the selected row color when flashing new
						// node. If selecting parent, use unselected row
						// color.
						endColor: args.selectParentPath ? '#fbfbfb' : '#ffeacf' //$NON-NLS-1$ //$NON-NLS-2$
					};

					var pathToFlash = args.selectParentPath || args.selectPath;
					tree_.flashNodeAtIdPath(pathToFlash, flashArgs);
				}
			});
		}

		deferredRefresh.addBoth(function() {
			refreshingMessage.hide();
		});

		return deferredRefresh;
	}

});

})();